from core import Controller
from kiss.views.core import JsonResponse

class RestController(object):
	"""
	Controller that creates REST API to your model.
	Pass model class to it and use url property and controller property in your urls settings.
	"""
	def __init__(self, model, id_regex=r"""(?P<id>\d+)"""):
		self.model = model
		self.id_regex = id_regex
		
	@property
	def url(self):
		return self.model.__name__.lower()
		
	@property
	def controller(self):
		return {
			"": RestListController(self.model),
			self.id_regex: RestShowController(self.model)
		}
	
	
def request_params_to_dict(params):
	result = {}
	for k,v in params.items():
		result[k] = v
	return result
	

class RestListController(Controller):
	def __init__(self, model):
		self.model = model
			
	def get(self, request):
		results = self.model.select()
		return JsonResponse(results)
	
	def post(self, request):
		result = self.model.create(**request_params_to_dict(request.form))
		return JsonResponse({"id": result.id}, status=201)

			
class RestShowController(Controller):
	def __init__(self, model):
		self.model = model
		
	def get(self, request):
		result = self.model.get(id=request.params["id"])
		return JsonResponse(result)
		
	def put(self, request):
		id = request.params["id"]
		self.model.update(**request_params_to_dict(request.form)).where(id=id).execute()
		return JsonResponse({"id": id})
	
	def delete(self, request):
		result = self.model.get(id=request.params["id"])
		result.delete_instance(recursive=True)
		return JsonResponse({"result": "ok"}, status=204)



import werkzeug.wrappers
from werkzeug.utils import cached_property
from werkzeug.utils import redirect
import jsonpickle
from peewee import Model, SelectQuery


class Request(werkzeug.wrappers.Request):
	"""
	Base request object inhereted from werkzeug Request.
	Added session object.
	"""
	def __init__(self, options, **argw):
		super(Request, self).__init__(options, **argw)
	
	@cached_property
	def session(self):
		return self.environ["session"]


class Response(werkzeug.wrappers.Response):
	"""
	Base response object inhereted from werkzeug Response.
	Text/html mimetype is default.
	"""
	def __init__(self, text, **argw):
		if "mimetype" not in argw:
			argw["mimetype"] = "text/html"
		super(Response, self).__init__(text, **argw)


class RedirectResponse(werkzeug.wrappers.Response):
	"""
	Response for redirect. Pass path and server will do 302 request.
	"""
	def __new__(cls, path):
		return redirect(path)
		
		
class JsonResponse(Response):
	"""
	Json response. Pass any object you want, JsonResponse converts it to json.
	"""
	def __init__(self, inp, **argw):
		inp = self.fix_attributes(inp)
		json_str = jsonpickle.encode(inp, unpicklable=False)
		super(JsonResponse, self).__init__(json_str, mimetype="application/json", **argw)
		
	def fix_attributes(self, inp):
		if isinstance(inp, SelectQuery) or (isinstance(inp, list)):
			inp = [self.fix_attributes(x) for x in list(inp)]
		elif isinstance(inp, Model):
			inp = self.fix_attributes_obj(inp)
		return inp
	
	def fix_attributes_obj(self, obj):
		new_dict = {}
		for key,value in obj.__dict__.items():
			value = self.fix_attributes(value)
			if key[:2] == "__":
				new_dict[key[2:]] = value
			elif key[:7] == "_cache_":
				new_dict[key[7:]] = value
			else:
				new_dict[key] = value
		obj.__dict__ = new_dict
		return obj



from core import Controller
from kiss.views.templates import TemplateResponse

class PageController(Controller):
	"""
	If you need just to show page, create PageController and pass to it your page and optional context.
	Use it like another controllers in urls settings of your app.
	"""
	def __init__(self, page, context={}):
		self.page = page
		self.context = context
		
	def get(self, request):
		return TemplateResponse(self.page, self.context)



from pev import Eventer

ApplicationStarted = 0
"""
Event when application is started	
"""

ApplicationStopped = 1
"""
Event when application is stopped
"""

BeforeDatabaseEngineConfiguration = 2
AfterDatabaseEngineConfiguration = 3



from kiss.views.templates import TemplateResponse
from kiss.views.core import RedirectResponse
from kiss.controllers.core import Controller
import requests
import json
from werkzeug.urls import url_decode
from putils.types import Dict, Regex
from putils.patterns import Singleton
from urlparse import urljoin
from werkzeug.urls import url_encode
	
	
class AuthBackend(object):
	"""
	Base oauth backend
	"""
	def get_code_url(self, request, options):
		params = {
			"client_id": options["client_id"],
			"redirect_uri": options["redirect_uri"],
			"scope": options["scope"],
			"response_type": "code",
			"approval_prompt": "force",
			"access_type": "offline"
		}
		return "%s?%s" % (options["authorization_uri"], url_encode(params))
		
	def get_access_token(self, request, options):
		params = {
			"client_id": options["client_id"],
			"client_secret": options["client_secret"],
			"grant_type": "authorization_code",
			"code": request.args["code"],
			"redirect_uri": options["redirect_uri"]
		}
		response = requests.post(options["get_token_uri"], params).text
		return self.prepare_access_token_response(response)
		
	def prepare_access_token_response(self, response):
		return json.loads(response)
		
	def get_user_info(self, request, options, access_token_result):
		self.access_token = access_token_result["access_token"]
		params = self.prepare_user_info_request_params(access_token_result)
		user_info_response = json.loads(requests.get("%s?%s" % (options["target_uri"], url_encode(params)), auth=self.auth).text)
		user_info_response = self.process_user_info_response(request, user_info_response)
		user_info_response["provider"] = request.params["backend"]
		user_info_response["access_token"] = params["access_token"]
		return user_info_response
				
	def prepare_user_info_request_params(self, access_token_result):
		return {"access_token": access_token_result["access_token"]}
		
	def process_user_info_response(self, request, user_info_response):
		result = {}
		result["id"] = user_info_response["id"]
		return result
		
	def auth(self, request):
		request.headers["Authorization"] = "Bearer %s" % self.access_token
		return request

		
class GoogleAuthBackend(AuthBackend):
	def process_user_info_response(self, request, user_info_response):
		result = {}
		result["id"] = user_info_response["id"]
		result["email"] = user_info_response["email"]
		result["name"] = user_info_response["name"]
		return result

	
class VkAuthBackend(AuthBackend):
	def prepare_user_info_request_params(self, access_token_result):
		return {"access_token": access_token_result["access_token"], "uids": access_token_result["user_id"], "fields": "uid, first_name, last_name, nickname, screen_name, sex, bdate, city, country, photo, photo_medium, photo_big"}
		
	def process_user_info_response(self, request, user_info_response):
		result = {}
		user_info_response = user_info_response["response"][0]
		result["id"] = user_info_response["uid"]
		result["name"] = "%s %s" % (user_info_response["first_name"], user_info_response["last_name"])
		return result

	
class FacebookAuthBackend(AuthBackend):
	def prepare_access_token_response(self, response):
		return url_decode(response)
		
	def process_user_info_response(self, request, user_info_response):
		result = {}
		result["id"] = user_info_response["id"]
		result["email"] = user_info_response["email"]
		result["name"] = user_info_response["name"]
		return result
		

class YandexAuthBackend(AuthBackend):	
	def process_user_info_response(self, request, user_info_response):
		result = {}
		result["id"] = user_info_response["id"]
		result["email"] = user_info_response["default_email"]
		result["name"] = user_info_response["real_name"]
		return result


class AuthManager(Singleton):
	"""
	Auth manager which calls appropriate backend
	"""	
	def __init__(self, opts):
		self.options = {
			"common": {
				"base_uri": "http://localhost:8080/auth/",
				"success_uri": "success/",
				"error_uri": "error/"
			},
			"google": {
				"authorization_uri": "https://accounts.google.com/o/oauth2/auth",
				"scope": "https://www.googleapis.com/auth/userinfo.profile https://www.googleapis.com/auth/userinfo.email",
				"get_token_uri": "https://accounts.google.com/o/oauth2/token",
				"redirect_uri": "google/callback",
				"target_uri": "https://www.googleapis.com/oauth2/v1/userinfo",
				"backend": GoogleAuthBackend()
			},
			"vk": {
				"authorization_uri": "http://api.vk.com/oauth/authorize",
				"scope": "",
				"get_token_uri": "https://api.vk.com/oauth/token",
				"redirect_uri": "vk/callback",
				"target_uri": "https://api.vk.com/method/users.get",
				"backend": VkAuthBackend()
			},
			"facebook": {
				"authorization_uri": "https://www.facebook.com/dialog/oauth",
				"scope": "email",
				"get_token_uri": "https://graph.facebook.com/oauth/access_token",
				"redirect_uri": "facebook/callback",
				"target_uri": "https://graph.facebook.com/me",
				"backend": FacebookAuthBackend()
			},
			"yandex": {
				"authorization_uri": "https://oauth.yandex.ru/authorize",
				"scope": "",
				"get_token_uri": "https://oauth.yandex.ru/token",
				"redirect_uri": "yandex/callback",
				"target_uri": "https://login.yandex.ru/info",
				"backend": YandexAuthBackend()
			}
		}
		self.options = Dict.merge(self.options, opts)
		base_uri = self.options["common"]["base_uri"]
		self.options["common"]["success_uri"] = urljoin(base_uri, self.options["common"]["success_uri"])
		self.options["common"]["error_uri"] = urljoin(base_uri, self.options["common"]["error_uri"])
		for backend, params in self.options.items():
			if "redirect_uri" in params:
				params["redirect_uri"] = urljoin(base_uri, params["redirect_uri"])
				
	def get_provider_url(self, request):
		current_options = self.options[request.params["backend"]]
		return current_options["backend"].get_code_url(request, current_options)
			
	def get_result_url(self, request):
		current_options = self.options[request.params["backend"]]
		access_token_result = current_options["backend"].get_access_token(request, current_options)
		if "access_token" not in access_token_result or not access_token_result["access_token"]:
			return self.options["common"]["error_uri"]
		user_info_response = current_options["backend"].get_user_info(request, current_options, access_token_result)
		return "%s?%s" % (self.options["common"]["success_uri"], url_encode(user_info_response))
		
class AuthController(object):
	"""
	Controller for social auth, use it in your url mappings
	"""
	def __new__(cls, opts):
		AuthManager(opts)
		return {
			Regex.string_url_regex("backend"): {
				"": StartAuthController,
				"callback": EndAuthController,
			}
		}

	
class StartAuthController(Controller):
	"""
	Controller which starts oauth flow.
	"""
	def get(self, request):
		if "success_uri" in request.args:
			AuthManager().options["common"]["success_uri"] = request.args["success_uri"]
		if "error_uri" in request.args:
			AuthManager().options["common"]["error_uri"] = request.args["error_uri"]
		return RedirectResponse(AuthManager().get_provider_url(request))
		

class EndAuthController(Controller):
	"""
	Controller which finishes oauth flow.
	"""
	def get(self, request):
		return RedirectResponse(AuthManager().get_result_url(request))




from putils.patterns import Singleton
from kiss.views.core import Response


class Controller(object):
	"""
	Base class of all controllers.
	"""
	def get(self, request):
		return Response("Method is not supported", status=405)
		
	def post(self, request):
		return Response("Method is not supported", status=405)
		
	def put(self, request):
		return Response("Method is not supported", status=405)
		
	def delete(self, request):
		return Response("Method is not supported", status=405)





BeforeControllerAction = 2
"""
Event before controller action will be executed.
You can change Request object before it will pass to controller action.
"""

AfterControllerAction = 3
"""
Event after controller action was executed.
You can change Response object before rendering.
"""



from putils.patterns import Singleton
from putils.filesystem import Dir
import mimetypes
import scss
from scss import Scss
from jsmin import jsmin
import os
import shutil
import traceback


class StaticCompiler(Singleton):
	"""
	Static files minifier.
	"""
	def __init__(self, path):
		self.css_parser = Scss()
		scss.LOAD_PATHS = path
		
	def compile_file(self, filepath):
		mimetype = mimetypes.guess_type(filepath)[0]
		return self.compile_text(self.get_content(filepath), mimetype)
		
	def compile_text(self, text, mimetype):
		result = ""
		if mimetype == "text/css":
			result = self.css_parser.compile(text)
		elif mimetype == "application/javascript":
			result = jsmin(text)
		else:
			result = text
		return result
		
	def get_content(self, file):
		return open(file).read()


class StaticBuilder(Singleton):
	"""
	Uses StaticCompiler to minify and compile js and css.
	"""
	def __init__(self, path):
		self.path = path
		self.compiler = StaticCompiler(self.path)
		
	def build(self):
		try:
			shutil.rmtree(self.path + "/build")
		except:
			pass
		try:
			Dir.walk(self.path, self.build_file)
		except:
			pass
		
	def build_file(self, file):
		new_path = self.path + "/build" + file.replace(self.path, "")
		result = self.compiler.compile_file(file)
		if result:
			try:
				os.makedirs(os.path.dirname(new_path))
			except:
				pass
			with open(new_path, "w") as f:
				f.write(result)



from kiss.views.core import Response
from kiss.core.application import Application


class Template(object):
	@staticmethod
	def text_by_path(path, context={}):
		return Application().options["views"]["templates_path"].get_template(path).render(context).encode("utf-8")
		
	@staticmethod
	def text_by_text(text, context={}):
		return Application().options["views"]["templates_path"].from_string(text).render(context).encode("utf-8")
		

class TemplateResponse(Response):
	"""
	Template response via Jinja2. Pass template path and context.
	"""
	def __init__(self, path, context={}, **argw):
		super(TemplateResponse, self).__init__(Template.text_by_path(path, context), **argw)
		
				
class TextResponse(Response):
	def __init__(self, text, context={}, **argw):
		super(TextResponse, self).__init__(Template.text_by_text(text, context), **argw)



from jinja2 import Environment, PackageLoader, ChoiceLoader
import re
from kiss.controllers.core import Controller
from putils.patterns import Singleton
from putils.types import Dict
from kiss.core.exceptions import *
from kiss.core.events import Eventer
import traceback
import events
import inspect


class Router(Singleton):
	"""
	Router implements unique hierarhical url mapping.
	Pass dictionary with mapping of regex and controller.
	"""
	def __init__(self, options):
		self.options = options
		self.eventer = Eventer()
		urls = Dict.flat_dict(self.options["urls"])
		new_urls = {}
		for k, v in urls.iteritems():
			if k[len(k)-2] == "/":
				k = k[:len(k)-2] + k[len(k)-1]
			k = re.compile(k)
			if inspect.isclass(v):
				new_urls[k] = v()
			else:
				new_urls[k] = v
		self.options["urls"] = new_urls
		if "templates_path" in self.options["views"]:
			tps = []
			if isinstance(self.options["views"]["templates_path"], list):
				for tp in self.options["views"]["templates_path"]:
					tps.append(PackageLoader(tp, ""))
			else:
				tps.append(PackageLoader(self.options["views"]["templates_path"], ""))
			self.options["views"]["templates_path"] = Environment(loader=ChoiceLoader(tps), extensions=self.options["views"]["templates_extensions"])
		
	def route(self, request):
		for re_url, controller in self.options["urls"].iteritems():
			path = request.path.lower()
			if path[len(path)-1] == "/":
				path = path.rstrip('/')
			mtch = re_url.match(path)
			if mtch:
				request.params = mtch.groupdict()
				try:
					self.eventer.publish(events.BeforeControllerAction, request)
					action = getattr(controller, request.method.lower())
					response = action(request)
					self.eventer.publish(events.AfterControllerAction, request, response)
					if not response:
						break
					return response
				except HTTPException, e:
					return self.get_err_page(e)
				except Exception, e:
					return self.get_err_page(InternalServerError(description=traceback.format_exc()))
		return self.get_err_page(NotFound(description="Not found %s" % request.url))
		
	def get_err_page(self, err):
		err_page = self.eventer.publish_and_get_result(err.code, err)
		if err_page:
			return err_page
		return err
		



from distutils.core import setup
try:
	from setuptools import setup
except:
	pass

setup(
    name = "kiss.py",
    version = "0.3.7",
    author = "Stanislav Feldman",
    description = ("MVC web framework in Python with Gevent, Jinja2, Werkzeug"),
    url = "http://stanislavfeldman.github.com/kiss.py/",
    keywords = "web framework gevent jinja2 werkzeug orm oauth socialauth vkontakte facebook google yandex",
    packages=[
    	"kiss", "kiss.controllers", "kiss.core", "kiss.views", "kiss.models"
    ],
    install_requires = ["gevent", "jinja2", "compressinja", "beaker", "werkzeug", "putils", "jsmin", "pyScss", "peewee", "jsonpickle", "pev"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Application Frameworks"
    ],
)



import gevent
import signal
from gevent import monkey; monkey.patch_all()
from gevent.wsgi import WSGIServer
from putils.patterns import Singleton
from putils.dynamics import Importer, Introspector
from putils.types import Dict
from kiss.controllers.router import Router
from kiss.views.core import Request, Response
from beaker.middleware import SessionMiddleware
from werkzeug.wsgi import SharedDataMiddleware
from kiss.core.events import Eventer, ApplicationStarted, ApplicationStopped, BeforeDatabaseEngineConfiguration, AfterDatabaseEngineConfiguration
from kiss.views.static import StaticBuilder
from kiss.models import Model
import logging


class Application(Singleton):
	"""
	Main class of your application.
	Pass options to constructor and all subsystems(eventer, router, db_engine) will be configured.
	"""
	def __init__(self, options):
		self.options = Application.init_options(options)
		self.options, self.eventer = Application.init_eventer(options)
		self.options, self.router = Application.init_router(options)
		self.eventer.publish(BeforeDatabaseEngineConfiguration, self)
		self.options, self.db_engine = Application.init_db(self.options)
		self.eventer.publish(AfterDatabaseEngineConfiguration, self)
		self.options, self.static_builder = Application.init_static(self.options)
		self.options, self.wsgi_app = Application.init_session(self.options, self.wsgi_app)
		self.options, self.wsgi_app = Application.init_static_server(self.options, self.wsgi_app)
		self.options, self.server = Application.init_server(self.options, self.wsgi_app)
	
	@staticmethod	
	def init_options(options):
		logging.basicConfig(level=logging.CRITICAL)
		default_options = {
			"application": {
				"address": "127.0.0.1",
				"port": 8080
			},
			"views": {
				"templates_path": "views.templates",
				"templates_extensions": ["compressinja.html.HtmlCompressor"],
				"static_path": "views.static",
				'session_type': "cookie",
				"session_auto": True,
				'session_cookie_expires': True,
				'session_encrypt_key':'sldk24j0jf09w0jfg24',
				'session_validate_key':';l[pfghopkqeq1234,fs'
			},
			"events": {}
		}
		return Dict.merge(default_options, options)
		
	@staticmethod	
	def init_eventer(options):
		return (options, Eventer(options["events"]))
		
	@staticmethod	
	def init_router(options):
		return (options, Router(options))
		
	@staticmethod
	def init_static(options):
		static_builder = None
		if "static_path" not in options["views"]:
			return (options, static_builder)
		try:
			options["views"]["static_path"] = Importer.module_path(options["views"]["static_path"])
		except:
			pass
		if options["views"]["static_path"]:
			static_builder = StaticBuilder(options["views"]["static_path"])
			static_builder.build()
		return (options, static_builder)
	
	@staticmethod			
	def init_db(options):
		if "models" not in options:
			return (options, None)
		db_engine_class = options["models"].pop("engine")
		db_name = options["models"].pop("database")
		db_engine = db_engine_class(db_name, **options["models"])
		db_engine.connect()
		db_engine.set_autocommit(True)
		for m in Introspector.all_subclasses(Model):
			m._meta.database = db_engine
		return (options, db_engine)
		
	@staticmethod
	def init_session(options, wsgi_app):
		session_options = {
			'session.type': options["views"]['session_type'],
			"session.auto": options["views"]["session_auto"],
			'session.cookie_expires': options["views"]['session_cookie_expires'],
			'session.encrypt_key': options["views"]['session_encrypt_key'],
			'session.validate_key': options["views"]['session_validate_key']
		}
		return (options, SessionMiddleware(wsgi_app, session_options, environ_key="session"))
		
	@staticmethod
	def init_static_server(options, wsgi_app):
		if "static_path" not in options["views"]:
			return (options, wsgi_app)
		else:
			return (options, SharedDataMiddleware(wsgi_app, {'/': options["views"]["static_path"] + "/build"}))
			
	@staticmethod
	def init_server(options, wsgi_app):
		kwargs = dict(filter(lambda item: item[0] not in ["address", "port"], options["application"].iteritems()))
		return (options, WSGIServer((options["application"]["address"], options["application"]["port"]), wsgi_app, **kwargs))
			
	def __del__(self):
		if self.db_engine:
			self.db_engine.close()
			
	def wsgi_app(self, options, start_response):
		request = Request(options)
		response = self.router.route(request)
		return response(options, start_response)
	
	def start(self):
		gevent.signal(signal.SIGTERM, self.stop)
		gevent.signal(signal.SIGINT, self.stop)
		self.eventer.publish(ApplicationStarted, self)
		self.server.serve_forever()
		
	def start_no_wait(self):
		self.eventer.publish(ApplicationStarted, self)
		self.server.start()
		
	def stop(self):
		self.eventer.publish(ApplicationStopped, self)
		self.server.stop()

